@@ -1504,6 +1504,7 @@ int main(int argc, char *argv[])
     int error = EOK;
     char *srcdir = NULL;
     char *rundir = NULL;
+    const char inidir[] = "/ini";
 
     if ((argc > 1) && (strcmp(argv[1], "-v") == 0)) verbose = 1;
 
@@ -1511,13 +1512,15 @@ int main(int argc, char *argv[])
 
     srcdir = getenv("srcdir");
     if(srcdir) {
-        rundir = malloc(strlen(srcdir)+4*sizeof(char));
-        sprintf(rundir, "%s/ini", srcdir);
+        rundir = malloc(strlen(srcdir) + sizeof(inidir));
+        sprintf(rundir, "%s%s", srcdir, inidir);
         if(chdir(rundir) != 0) {
             error = errno;
+            free(rundir);
             printf("Failed to change directory, error %d\n", error);
             return error;
         }
+        free(rundir);
     }
 
     if ((error = basic_test()) ||
